/* $Id: protocol.c,v 1.75 2006/05/14 16:34:52 darkjames Exp $ */

/*
 *  (C) Copyright 2003 Wojtek Kaniewski <wojtekka@irc.pl>
 *		  2004 Piotr Kupisiewicz <deli@rzepaknet.us>
 *		  2004 Adam Mikuta <adammikuta@poczta.onet.pl>
 *		  2005 Leszek Krupiski <leafnode@wafel.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ekg2-config.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>

#include "commands.h"
#include "log.h"
#include "emoticons.h"
#include "msgqueue.h"
#include "objects.h"
#include "protocol.h"
#include "stuff.h"
#include "userlist.h"
#include "themes.h"
#include "windows.h"
#include "xmalloc.h"

static int auto_find_limit = 100; /* counter of persons who we were looking for when autofind */
list_t dccs = NULL;

/*
 * protocol_init()
 *
 * rejestruje wszystkie zapytania zwizane z protokoem, ktre bd
 * wysyane przez pluginy.
 */
void protocol_init()
{
	query_connect(NULL, "protocol-status", protocol_status, NULL);
	query_connect(NULL, "wcs_protocol-status", protocol_status, (void *) 1);
	query_connect(NULL, "protocol-message", protocol_message, NULL);
	query_connect(NULL, "wcs_protocol-message", protocol_message, (void *) 1);
	query_connect(NULL, "protocol-message-ack", protocol_message_ack, NULL);

	query_connect(NULL, "protocol-connected", protocol_connected, NULL);
	query_connect(NULL, "wcs_protocol_disconnected", protocol_disconnected, (void *) 1);
	query_connect(NULL, "protocol-disconnected", protocol_disconnected, NULL);
}


/*
 * protocol_reconnect_handler()
 *
 * obsuga timera reconnectu.
 */
static TIMER(protocol_reconnect_handler)
{
	char *session = (char*) data;
	session_t *s = session_find(session);

	if (type == 1) {
		xfree(session);
		return 0;
	}

        if (!s || session_connected_get(s) == 1)
                return -1;

	debug("reconnecting session %s\n", session);

	command_exec(NULL, s, TEXT("/connect"), 0);
	return 0;
}

/*
 * protocol_disconnect()
 *
 * obsuga zerwanego poczenia.
 */
int protocol_disconnected(void *data, va_list ap)
{
	char *session	= *(va_arg(ap, char **));
	char *reason	= *(va_arg(ap, char **));
	int type	= *(va_arg(ap, int*));

#if USE_UNICODE
	if (data == (void *) 1) {
		int ret;
		char *__session = wcs_to_normal((CHAR_T *) session);
		char *__reason	= wcs_to_normal((CHAR_T *) reason);

		ret = query_emit(NULL, "protocol-disconnected", &__session, &__reason, &type);

		free_utf(__session);
		free_utf(__reason);
		return ret;
	}
#endif
	
	userlist_clear_status(session_find(session), NULL);

	switch (type) {
		case EKG_DISCONNECT_NETWORK:
		case EKG_DISCONNECT_FAILURE:
		{
			int tmp;
			session_t *s = session_find(session);

			if (!s)
				break;

			if ((tmp = session_int_get(s, "auto_reconnect")) && tmp != -1)
				timer_add(plugin_find_uid(s->uid), "reconnect", tmp, 0, protocol_reconnect_handler, xstrdup(session));

			if (type == EKG_DISCONNECT_NETWORK)
				print("conn_broken", session_name_n(session));
			else
				print("conn_failed", reason, session_name_n(session));
			break;
		}

		case EKG_DISCONNECT_USER:
			if (reason)
				print("disconnected_descr", reason, session_name_n(session));
			else
				print("disconnected", session_name_n(session));
			break;

		case EKG_DISCONNECT_FORCED:
			print("conn_disconnected", session_name_n(session));
			break;
			
		case EKG_DISCONNECT_STOPPED:
			print("conn_stopped", session_name_n(session));
			break;

		default:
			wcs_print("generic_error", TEXT("protocol_disconnect internal error, report to authors"));
			break;
	}

	return 0;
}

/*
 * protocol_connect()
 *
 * obsuga udanego poczenia.
 */
int protocol_connected(void *data, va_list ap)
{
	char **session = va_arg(ap, char**);
	session_t *s = session_find(*session);
	const char *descr = session_descr_get(s);
	
        ekg_update_status(s);

	if (descr)
		print("connected_descr", descr, session_name(s));
	else
		print("connected", session_name(s));

	if (!msg_queue_flush(*session))
		print("queue_flush", session_name(s));

	return 0;
}

/*
 * protocol_status()
 *
 * obsuga zapytania "protocol-status" wysyanego przez pluginy protokow.
 */
int protocol_status(void *data, va_list ap)
{
	char **__session	= va_arg(ap, char**), *session = *__session;
	char **__uid		= va_arg(ap, char**), *uid = *__uid;
	char *status		= *(va_arg(ap, char**));
	char **__descr		= va_arg(ap, char**), *descr = *__descr;
	char *host		= *(va_arg(ap, char**));
	int port		= *(va_arg(ap, int*));
	time_t when		= *(va_arg(ap, time_t*));

	userlist_t *u;
	session_t *s;
	int ignore_level;
        int ignore_status, ignore_status_descr, ignore_events, ignore_notify;
#if USE_UNICODE
	if (data == (void *) 1) {
		char *ssession	= wcs_to_normal( (CHAR_T *) session);
		char *suid	= wcs_to_normal( (CHAR_T *) uid);
		char *sstatus	= wcs_to_normal( (CHAR_T *) status);
		char *sdescr	= wcs_to_normal( (CHAR_T *) descr);
		char *shost	= wcs_to_normal( (CHAR_T *) host);
		int ret;

		ret = query_emit(NULL, "protocol-status", &ssession, &suid, &sstatus, &sdescr, &shost, &port, &when);

		free_utf(ssession);
		free_utf(suid);
		free_utf(sstatus);
		free_utf(sdescr);
		free_utf(shost);
		return ret;
	}
#endif
	if (!(s = session_find(session)))
		return 0;

	/* we are checking who user we know */
	if (!(u = userlist_find(s, uid))) {
		if (config_auto_user_add) {
			if (!(u=userlist_add(s, uid, uid)))
				return 0;
		} else {
			return 0;
		}
	}
	ignore_level = ignored_check(s, uid);

	ignore_status = ignore_level & IGNORE_STATUS;
	ignore_status_descr = ignore_level & IGNORE_STATUS_DESCR;
	ignore_events = ignore_level & IGNORE_EVENTS;
	ignore_notify = ignore_level & IGNORE_NOTIFY;

	/* zapisz adres IP i port */
	u->ip = (host) ? inet_addr(host) : 0;
	u->port = port;

	if (host)
		u->last_ip = inet_addr(host);
	if (port)
		u->last_port = port;

	/* jeli te same stany... */
	if (!xstrcasecmp(status, u->status)) {
		/* ...i nie ma opisu, ignoruj */
		if (!descr && !u->descr)
			return 0;
	
		/* ...i te same opisy, ignoruj */
		if (descr && u->descr && !xstrcmp(descr, u->descr))
			return 0;
	}

	/* jeli kto nam znika, zapamitajmy kiedy go widziano */
	if (xstrcasecmp(u->status, EKG_STATUS_NA) && !xstrcasecmp(status, EKG_STATUS_NA))
		u->last_seen = when ? when : time(NULL);

	/* XXX doda events_delay */
	
	/* jeli dostpny lub zajty, dopisz to taba. jeli niedostpny, usu */
	if (!xstrcasecmp(status, EKG_STATUS_AVAIL) && config_completion_notify && u->nickname)
		tabnick_add(u->nickname);
	if (!xstrcasecmp(status, EKG_STATUS_AWAY) && (config_completion_notify & 4) && u->nickname)
		tabnick_add(u->nickname);
	if (!xstrcasecmp(status, EKG_STATUS_NA) && (config_completion_notify & 2) && u->nickname)
		tabnick_remove(u->nickname);


	/* jeli mao wana zmiana stanu... */
	if ((session_int_get(s, "display_notify") == 2 || (session_int_get(s, "display_notify") == -1 && config_display_notify == 2)) && xstrcasecmp(u->status, EKG_STATUS_NA)) {
		/* jeli na zajty, ignorujemy */
		if (!xstrcasecmp(status, EKG_STATUS_AWAY))
			goto notify_plugins;

		/* jeli na dostpny, ignorujemy */
		if (!xstrcasecmp(status, EKG_STATUS_AVAIL))
			goto notify_plugins;
	}

	/* ignorowanie statusu - nie wywietlamy, ale pluginy niech robi co chc */
        if (ignore_status)
		goto notify_plugins;

	/* nie zmieni si status, zmieni si opis */
	if (ignore_status_descr && !xstrcmp(status, u->status) && xstrcmp(descr, u->descr))
		goto notify_plugins;

	/* daj zna dwikiem... */
	if (config_beep && config_beep_notify)
		query_emit(NULL, "ui-beep");

	/* ...i muzyczk */
	if (config_sound_notify_file)
		play_sound(config_sound_notify_file);

        /* wywietla na ekranie? */
        if (!session_int_get(s, "display_notify")) 
                goto notify_plugins;

	if (!config_display_notify && session_int_get(s, "display_notify") == -1)
		goto notify_plugins;

	/* poka */
	if (u->nickname) {
		char format[100];
		if (ignore_status_descr) {
			snprintf(format, sizeof(format), "status_%s%s", status, "");
		} else {
			snprintf(format, sizeof(format), "status_%s%s", status, (descr) ? "_descr" : "");
		}
		print_window(u->nickname, s, 0, format, format_user(s, uid), (u->first_name) ? u->first_name : u->nickname, session_name(s), descr);
	}

notify_plugins:
	if (xstrcasecmp(u->status, EKG_STATUS_NA)) {
	        xfree(u->last_status);
	        u->last_status = xstrdup(u->status);
	        xfree(u->last_descr);
	        u->last_descr = xstrdup(u->descr);
	}

	if (!xstrcasecmp(u->status, EKG_STATUS_NA) && xstrcasecmp(status, EKG_STATUS_NA) && !ignore_events)
		query_emit(NULL, "event_online", __session, __uid);

	if (!ignore_status) {
		xfree(u->status);
		u->status = xstrdup(status);
	}

	if (xstrcasecmp(u->descr, descr) && !ignore_events)
		query_emit(NULL, "event_descr", __session, __uid, __descr);

	if (!ignore_status && !ignore_status_descr) {
		xfree(u->descr);
		u->descr = xstrdup(descr);
		u->status_time = time(NULL);
	}
	
	query_emit(NULL, "userlist-changed", __session, __uid);

	if (!xstrcasecmp(status, EKG_STATUS_AVAIL) && !ignore_events)
		query_emit(NULL, "event_avail", __session, __uid);
	if (!xstrcasecmp(status, EKG_STATUS_AWAY) && !ignore_events)
                query_emit(NULL, "event_away", __session, __uid);
        if (!xstrcasecmp(status, EKG_STATUS_NA) && !ignore_events)
                query_emit(NULL, "event_na", __session, __uid);

	return 0;
}

/*
 * message_print()
 *
 * wywietla wiadomo w odpowiednim oknie i w odpowiedniej postaci.
 *
 * zwraca target
 */
char *message_print(const char *session, const char *sender, const char **rcpts, const char *__text, const uint32_t *format, time_t sent, int class, const char *seq, int dobeep, int secure)
{
	char *class_str, timestamp[100], *t = NULL, *text = xstrdup(__text), *emotted = NULL;
	const char *target = sender, *user;
        time_t now;
	session_t *s = session_find(session);
        struct conference *c = NULL;
	int empty_theme = 0;

	if (class & EKG_NO_THEMEBIT) {
		empty_theme = 1;
		class &= ~EKG_NO_THEMEBIT;
	}

	switch (class) {
		case EKG_MSGCLASS_SENT:
			class_str = "sent";
			target = (rcpts) ? rcpts[0] : NULL;
			break;
		case EKG_MSGCLASS_CHAT:
			class_str = "chat";
			break;
		case EKG_MSGCLASS_SYSTEM:
			class_str = "system";
			target = "__status";
			break;
		case EKG_MSGCLASS_SENT_CHAT:
			class_str = "sent";
                        target = (rcpts) ? rcpts[0] : NULL;
        	        break;
		default:
			class_str = "message";
	}

	/* dodajemy kolorki do tekstu */
	if (format) {
		string_t s = string_init("");
		const char *attrmap;
		int i;

		if (config_display_color_map && xstrlen(config_display_color_map) == 8)
			attrmap = config_display_color_map;
		else
			attrmap = "nTgGbBrR";

		for (i = 0; text[i]; i++) {
			if (i == 0 || format[i] != format[i - 1]) {
				char attr = 'n';

				if ((format[i] & EKG_FORMAT_COLOR)) {
					attr = color_map(format[i] & 0x0000ff, (format[i] & 0x0000ff00) >> 8, (format[i] & 0x00ff0000) >> 16);
					if (attr == 'k')
						attr = 'n';
				}

				if ((format[i] & 0xfe000000)) {
					uint32_t f = (format[i] & 0xff000000);

					if ((format[i] & EKG_FORMAT_COLOR))
						attr = toupper(attr);
					else
						attr = attrmap[(f >> 25) & 7];
				}

				string_append_c(s, '%');
				string_append_c(s, attr);
			}

			if (text[i] == '%')
				string_append_c(s, '%');
			
			string_append_c(s, text[i]);
		}

		xfree(text);
		text = format_string(s->str);
		string_free(s, 1);
	}

	/* wyznaczamy odstp czasu midzy wysaniem wiadomoci, a chwil
	 * obecn, eby wybra odpowiedni format timestampu. */
	{
		char tmp[100], *timestamp_type;
	        struct tm *tm_now, *tm_msg;

		now = time(NULL);
		tm_now = localtime(&now);
		tm_msg = localtime(&sent);

		if (sent - config_time_deviation <= now && now <= sent + config_time_deviation)
			timestamp_type = "timestamp_now";
		else if (tm_now->tm_yday == tm_msg->tm_yday)
			timestamp_type = "timestamp_today";
		else	timestamp_type = "timestamp";

		snprintf(tmp, sizeof(tmp), "%s_%s", class_str, timestamp_type);
		if (!strftime(timestamp, sizeof(timestamp), format_find(tmp), tm_msg)
				&& xstrlen(format_find(tmp))>0)
			xstrcpy(timestamp, "TOOLONG");
	}

	/* if there is a lot of recipients, conference should be made */
	{
		int recipients_count = array_count((char **) rcpts);

		if (xstrcmp(class_str, "sent") && recipients_count > 0) {
			c = conference_find_by_uids(s, sender, rcpts, recipients_count, 0);

	                if (!c) {
	                        string_t tmp = string_init(NULL);
	                        int first = 0, i;
	
	                        for (i = 0; i < recipients_count; i++) {
	                                if (first++)
	                                        string_append_c(tmp, ',');
	
	                                string_append(tmp, rcpts[i]);
	                        }
	
	                        string_append_c(tmp, ' ');
	                        string_append(tmp, sender);
	
	                        c = conference_create(s, tmp->str);
	
	                        string_free(tmp, 1);
	                }

	                if (c) {
	                        target = c->name;
				class_str = "conference";
			}
		}
	}

	/* daj zna dwikiem i muzyczk */
	if (class == EKG_MSGCLASS_CHAT) {

		if (config_beep && config_beep_chat && dobeep)
			query_emit(NULL, "ui-beep");
	
		if (config_sound_chat_file && dobeep)
			play_sound(config_sound_chat_file);

	} else if (class == EKG_MSGCLASS_MESSAGE) {

		if (config_beep && config_beep_msg && dobeep)
			query_emit(NULL, "ui-beep");
		if (config_sound_msg_file && dobeep)
			play_sound(config_sound_chat_file);

	} else if (class == EKG_MSGCLASS_SYSTEM && config_sound_sysmsg_file)
			play_sound(config_sound_sysmsg_file);
	
        if (config_last & 3 && xstrcasecmp(class_str, "sent")) 
	        last_add(0, sender, now, sent, text);
	
	user = xstrcasecmp(class_str, "sent") ? format_user(s, sender) : session_format_n(sender);

	if (config_emoticons && text)
		emotted = emoticon_expand(text);

	if (empty_theme)
		class_str = "empty";

	print_window(target, s, (class == EKG_MSGCLASS_CHAT || class == EKG_MSGCLASS_SENT_CHAT), class_str, user, timestamp, (emotted) ? emotted : text, (!xstrcasecmp(class_str, "sent")) ? session_alias_uid(s) : get_nickname(s, sender), (!xstrcasecmp(class_str, "sent")) ? s->uid : get_uid(s, sender), (secure) ? format_string(format_find("secure")) : "");

	xfree(text);
	xfree(t);
	xfree(emotted);
	return xstrdup(target);
}

/*
 * protocol_message()
 */
int protocol_message(void *data, va_list ap)
{
	char *session	= *(va_arg(ap, char**));
	char *uid	= *(va_arg(ap, char**));
	char **rcpts	= *(va_arg(ap, char***));
	char *text	= *(va_arg(ap, char**));
	uint32_t *format= *(va_arg(ap, uint32_t**));
	time_t sent	= *(va_arg(ap, time_t*));
	int class	= *(va_arg(ap, int*));
	char *seq	= *(va_arg(ap, char**));
	int dobeep	= *(va_arg(ap, int*));
	int secure	= *(va_arg(ap, int*));

	session_t *session_class = session_find(session);
	userlist_t *userlist = userlist_find(session_class, uid);
	char *target = NULL;
	int empty_theme = 0;
	int our_msg;

#if USE_UNICODE
	if (data == (void *) 1) {
		char *ssession	= wcs_to_normal( (CHAR_T *) session);
		char *suid	= wcs_to_normal( (CHAR_T *) uid);
		char *stext	= wcs_to_normal( (CHAR_T *) text);
		char *sseq	= wcs_to_normal( (CHAR_T *) seq);
		char **srcpts   = wcs_array_to_str( (CHAR_T **) rcpts);
		int ret;

		ret = query_emit(NULL, "protocol-message", &ssession, &suid, &srcpts, &stext, &format, &sent, &class, &sseq, &dobeep, &secure);

		free_utf(ssession);
		free_utf(suid);
		free_utf(stext);
		free_utf(sseq);
		array_free(srcpts);
		return ret;
	}
#endif

	if (ignored_check(session_class, uid) & IGNORE_MSG)
		return -1;

	/* display blinking */
	if (config_display_blinking && userlist && class != EKG_MSGCLASS_SENT && class != EKG_MSGCLASS_SENT_CHAT && (!rcpts || !rcpts[0])) {
		if (config_make_window && xstrcmp(get_uid(session_class, window_current->target), get_uid(session_class, uid))) 
			userlist->blink = 1;
		else if (!config_make_window) {
			window_t *w;

			/*
			 * now we are checking if there is some window with query for this
			 * user 
			 */
			w = window_find_s(session_class, uid);

			if (!w && window_current->id != 1)
				userlist->blink = 1; 
			if (w && window_current->id != w->id)
				userlist->blink = 1;
		}
	}
	
	if (class & EKG_NO_THEMEBIT) {
		class &= ~EKG_NO_THEMEBIT;
		empty_theme = 1;
	}
	our_msg = (class == EKG_MSGCLASS_SENT || class == EKG_MSGCLASS_SENT_CHAT);

	/* there is no need to decode our messages */
	if (!our_msg) {
                char *___session = xstrdup(session);
                char *___sender = xstrdup(uid);
                char *___message = xstrdup(text);
                int ___decrypted = 0;

                query_emit(NULL, "message-decrypt", &___session, &___sender, &___message, &___decrypted, NULL);

                if (___decrypted) {
                        text = ___message;
                        ___message = NULL;
                        secure = 1;
                }

                xfree(___session);
                xfree(___sender);
                xfree(___message);
	}

	if (our_msg)	query_emit(NULL, "protocol-message-sent", &session, &(rcpts[0]), &text);
	else		query_emit(NULL, "protocol-message-received", &session, &uid, &rcpts, &text, &format, &sent, &class, &seq, &secure);

	query_emit(NULL, "protocol-message-post", &session, &uid, &rcpts, &text, &format, &sent, &class, &seq, &secure);

	/* show it ! */
	if (!(our_msg && !config_display_sent)) {
		if (empty_theme)
			class |= EKG_NO_THEMEBIT;
	        target = message_print(session, uid, (const char**) rcpts, text, format, sent, class, seq, dobeep, secure);
	}

        /* jeeli nie mamy podanego uid'u w licie kontaktw to trzeba go dopisa do listy dopenianych */
	if (!userlist) 
		tabnick_add(uid);

        if (!userlist && xstrcasecmp(session_class->uid, uid) && session_int_get(session_class, "auto_find") >= 1) {
                list_t l;
                int do_find = 1, i;

                for (l = autofinds, i = 0; l; l = l->next, i++) {
                        char *d = l->data;

                        if (!xstrcmp(d, uid)) {
                                do_find = 0;
                                break;
                        }
                }

                if (do_find) {
                        if (i == auto_find_limit) {
                                debug("// autofind reached %d limit, removing the oldest uin: %d\n", auto_find_limit, *((char *)autofinds->data));
                                list_remove(&autofinds, autofinds->data, 1);
                        }

                        list_add(&autofinds, (void *) uid, xstrlen(uid) + 1);

                        command_exec_format(target, session_class, 0, TEXT("/find %s"), uid);
                }
        }

	xfree(target);
	return 0;
}

/*
 * protocol_message_ack()
 */
int protocol_message_ack(void *data, va_list ap)
{
	CHAR_T *__session	= *(va_arg(ap, CHAR_T **));
	CHAR_T *__rcpt		= *(va_arg(ap, CHAR_T **));
	CHAR_T *__seq		= *(va_arg(ap, CHAR_T **));
	CHAR_T *__status	= *(va_arg(ap, CHAR_T **));

	char *session	= wcs_to_normal(__session);
	char *rcpt	= wcs_to_normal(__rcpt);
	char *seq	= wcs_to_normal(__seq);

	userlist_t *u = userlist_find(session_find(session), rcpt);
	const char *target = (u && u->nickname) ? u->nickname : rcpt;
	int display = 0;
	CHAR_T format[100];
#if USE_UNICODE
	swprintf(format, sizeof(format)/sizeof(CHAR_T), TEXT("ack_%ls"), __status);
#else
	snprintf(format, sizeof(format), "ack_%s", __status);
#endif

	msg_queue_remove_seq(seq);
	
	if (config_display_ack == 1)
		display = 1;

	if (!xwcscmp(__status, EKG_ACK_DELIVERED) && config_display_ack == 2)
		display = 1;

	if (!xwcscmp(__status, EKG_ACK_QUEUED) && config_display_ack == 3)
		display = 1;

	if (display)
		print_window(target, session_find(session), 0, wcs_to_normal(format), format_user(session_find(session), rcpt));

	free_utf(session);
	free_utf(rcpt);
	free_utf(seq);

	return 0;
}

dcc_t *dcc_add(const char *uid, dcc_type_t type, void *priv)
{
	dcc_t *d = xmalloc(sizeof(dcc_t));
	int id = 1, id_ok = 0;
	list_t l;

	memset(d, 0, sizeof(dcc_t));

	d->uid = xstrdup(uid);
	d->type = type;
	d->priv = priv;
	d->started = time(NULL);

	do {
		id_ok = 1;

		for (l = dccs; l; l = l->next) {
			dcc_t *D = l->data;
	
			if (D->id == id) {
				id++;
				id_ok = 0;
				break;
			}
		}
	} while (!id_ok);

	d->id = id;

	list_add(&dccs, d, 0);

	return d;
}

int dcc_close(dcc_t *d)
{
	if (!d)
		return -1;

	if (d->close_handler)
		d->close_handler(d);

	xfree(d->uid);
	xfree(d->filename);

	list_remove(&dccs, d, 1);

	return 0;
}

PROPERTY_MISC(dcc, close_handler, dcc_close_handler_t, NULL)
PROPERTY_STRING(dcc, filename)
PROPERTY_INT(dcc, offset, int)
PROPERTY_INT(dcc, size, int)
PROPERTY_STRING_GET(dcc, uid)
PROPERTY_INT_GET(dcc, id, int)
PROPERTY_PRIVATE(dcc)
PROPERTY_INT_GET(dcc, started, time_t)
PROPERTY_INT(dcc, active, int)
PROPERTY_INT(dcc, type, dcc_type_t)


/*
 * Local Variables:
 * mode: c
 * c-file-style: "k&r"
 * c-basic-offset: 8
 * indent-tabs-mode: t
 * End:
 * vim: sts=0 noexpandtab sw=8
 */
